/*
	httpclient.h
	part of client to wininet http facilities
	
	version 2.2
	
	2.2 added 'simple' HTTP client
	2.1 added xTPostData structure
	2.0 added cookie support
	1.0 post method created
	0.1 initial version
	
	Written by Povilas Kanapickas [exec]

	This software comes "AS IS" therefore the author is NOT responsible for any 
	losses or damage it can cause.

	Published under GPL (see http://www.gnu.org/gpl)
*/

#ifndef HTTPCLIENT_H
#define HTTPCLIENT_H

#define simplehttp
#ifdef simplehttp
#include "xmlfile.h"
#endif

#include "windows.h"
#include "wininet.h"
#include <vector>
#include <string>

using namespace std;

typedef unsigned long ulong;

/* class to post data easily - just add item/value pairs to the
	structure and call wGetData() to get data in format 
	item=value&item2=value2
*/

class xTPostData	{
	private:
		string sData;
		bool bCleared;
	public:
		xTPostData()	{bCleared = true;};
		void wAddData (const char *psName, const char *psValue);
		void wAddData (const char *psName, int iValue);
		const char *wGetData ()	{ return sData.c_str(); bCleared = true;}
	};
			
/*================================*/

class xTHeader	{
	public:
		char *sHeader;
		char *sValue;
	
		xTHeader()	{
			sHeader = NULL;
			sValue = NULL;
			};
			
		~xTHeader()	{
			delete[] sHeader;
			delete[] sValue;
			};
	};
/*================================*/
/* class to add request headers easily - just add header/value pairs
	to the structure and call wGetHeaders() to get data in format 
	item: value \r\n item2: value2 (uses CR/LF to separate headers)
*/
typedef vector<xTHeader*> xTHeaderList;

class xTHeaders	{
	private:
		xTHeaderList xHeaders;
			 
	public:
		void wAddHeader(xTHeader *header);
		void wAddHeader(const char *psHeader, const char *psValue);
		const char *wGetHeaders();
		~xTHeaders();
	};
/*================================*/
/* class to connect to internet. Connectting takes place in constructor*/
typedef enum {get, put, post}xTMethod;

class xTHTTPConnection	{
	private:
		bool bConnected;
		HINTERNET xConnection;
		char *sClient;
	public:
		
		HINTERNET wConnection()	{
			return xConnection;
			}
		
		bool wConnected()	{
			return bConnected;
			}
		
		xTHTTPConnection();
		~xTHTTPConnection()	{
			if (bConnected)
				InternetCloseHandle (xConnection);
			}
	};
//================================================
/* class to connect to server. Connecting takes place in constructor.
*/
class xTHTTPSession	{
	private:
		HINTERNET xSession;
		bool bConnected;
		char *psCookie;
	public:
		
		HINTERNET wSession()	{
			return xSession;
			}
		
		bool wConnected()	{
			return bConnected;
			}
			
		xTHTTPSession (xTHTTPConnection *connection, const char *server);
		
		~xTHTTPSession()	{
			if (bConnected)
				InternetCloseHandle (xSession);
			}
		void wSetCookie (char *sCookie)	{
			if (psCookie) delete psCookie;
			psCookie = sCookie;
			}
		char *wGetCookie()	{return psCookie;};
	};
//===============================================
/* class to get resource from server
*/
class xTHTTPResource	{
	private:
		HINTERNET xRequest;
		bool bConnected;
		
	public:
		
		// creates a request and sends it to server
		xTHTTPResource(	xTHTTPSession *session, 
						xTMethod verb, 
						const char *sURL, 
						const char *sReferer, 
						xTHeaders *headers, 
						const char *sData);
						
		// method to get data from server. length is buffer size in chars. 
		// Returns number of chars read.
		long wReadData(ulong length, char *buf);
		
		~xTHTTPResource()	{
			if (bConnected) InternetCloseHandle (xRequest);
			}
		bool wConnected()	{
			return bConnected;
			}
	};
//===============================================
#ifdef simplehttp
#define SIMPLEHTTPBUFLEN 102401
/* class to interface with server easily. Reads all the data into xTXMLFile structure
*/
class xTSimpleHTTPClient	{
	private:
		xTHTTPConnection 	*hpxConnection;
		xTHTTPSession 		*hpxSession;
		xTHTTPResource 		*hpxRes;
		
		bool hbConnected;
		char *hpsLastServer;
		char *hpsLastFile;
		char *hpsBuffer;
	public:
		
		xTSimpleHTTPClient();
		~xTSimpleHTTPClient();
		int wConnect(const char *psServer);
		int wGet(xTXMLFile *pxHTML, const char *psFile);
		int wPost(xTXMLFile *pxHTML, const char *psFile, const char *psData);
		int wGet(xTXMLFile *pxHTML, const char *psFile, const char *psData);
		};	
#endif		
		
#endif
