/* 	xmlfile.cpp
Part of xml file parser

	version 2.2
	
	Written by Povilas Kanapickas [exec]

	This software comes "AS IS" therefore the author is NOT responsible for any 
	losses or damage it can cause.

	Published under GPL (see http://www.gnu.org/gpl)
*/

#include "xmlfile.h"
#include "string.h"

void wStrCopyNoWhiteSpace (char *psDest, const char *psSrc)	{
	strcpy (psDest, psSrc);
	int iLen = strlen (psDest);
	while (*(psDest+iLen) < 31) if (!iLen) break; 
									else iLen--;
	*(psDest+iLen+1) = 0;
	}
	

xTXMLElement::xTXMLElement(xTXMLElement *pxParent, xTType eType, const char* psName)	{ 	
	static char a = 0;
	psNul = &a;
	heType = eType;
	hpsName = new char[strlen(psName) + 1];
	wStrCopyNoWhiteSpace (hpsName, psName);
	
	switch (heType)	{
		
		case node:	{
			hpxAttributes = NULL;
			hpxChilds = NULL;
			hpsValue = NULL;
			break;
			}
			
		case attr:	{
			hpxAttributes = NULL;
			hpxChilds = NULL;
			hpsValue = psNul;
			break;
			}
		
		default:	{ //for text, special
			hpxAttributes = NULL;
			hpxChilds = NULL;
			hpsValue = NULL;
			break;
			}
		}
	hpxCurrentChild = NULL;
	hpxCurrentAttr = NULL;
	hpxParent = pxParent;
	hbOnDelete = false;
	}
//============================================
//to create root element
xTXMLElement::xTXMLElement()	{
	heType = node;
	hpsName = "root";
	hpxAttributes = NULL;
	hpxChilds = NULL;
	hpsValue = NULL;
	hpxParent = NULL;
	hbOnDelete = false;
	}
//============================================
//deletes an element
xTXMLElement::~xTXMLElement()	{
	hbOnDelete = true;
	switch (heType)	{
	
		case node:	{
			xTXMLElementListItem *iter;
			while (hpxAttributes)	{
				delete hpxAttributes->pxElement;
				
				iter = hpxAttributes;
				hpxAttributes = hpxAttributes->pxNext;
				delete iter;
				}
			while (hpxChilds)	{
				delete hpxChilds->pxElement;
				
				iter = hpxChilds;
				hpxChilds = hpxChilds->pxNext;
				delete iter;
				}
			delete hpsName;
			break;
			}
			
		case attr:	{
			if (hpsValue != psNul) 
				delete hpsValue;
			if (hpsName != psNul) 
				delete hpsName;
			break;
			}
			
		default: {
			if (hpsName != psNul) 
				delete hpsName;
			return;}
		}
	if (hpxParent) hpxParent->wRemove(this);
	}
//=============================================
int xTXMLElement::wGetInt()	{ 
	int iNum = 0;
	sscanf (hpsName, "%d", &iNum);
	return iNum;
	}
//==============================================
xTXMLElement *xTXMLElement::wSearchNext (xTType eType, const char *sName)	{
	#ifdef DEBUGSRC
	{FILE *pxFR = fopen (DEBUGSRCFILE, "a");
	fprintf (pxFR, "%s\n", hpsName);
	fclose (pxFR);}
	#endif
	
	if (strcmp (sName, hpsName) == 0) if (heType == eType) return (xTXMLElement*)1; 
	
	xTXMLElement *pxResult;
	while (hpxCurrentChild){
		pxResult = hpxCurrentChild->pxElement->wSearchNext(eType, sName);
		if ((int)pxResult == 1) {
			pxResult = hpxCurrentChild->pxElement;
			hpxCurrentChild = hpxCurrentChild->pxNext;
			return pxResult;
			}
		if (pxResult) 	{return pxResult;}
		hpxCurrentChild = hpxCurrentChild->pxNext;
		}
	hpxCurrentChild = hpxChilds;
	return NULL;
	}
//===============================================
xTXMLElement *xTXMLElement::wSearchNextNameAttr (const char *sName, const char *sAttrName, const char *sAttrValue)	{
	#ifdef DEBUGSRC
	{FILE *pxFR = fopen (DEBUGSRCFILE, "a");
	fprintf (pxFR, "%s\n", hpsName);
	fclose (pxFR);}
	#endif
	
	const char *val;
	if (strcmp (sName, hpsName) == 0)	{
		if (heType != node) return NULL;
		val = wSearchAttr(sAttrName);
		if (val)
			if(strcmp (val, sAttrValue) == 0) return (xTXMLElement*)1; 
		}
	
	xTXMLElement *pxResult;
	while (hpxCurrentChild){
		pxResult = hpxCurrentChild->pxElement->wSearchNextNameAttr(sName, sAttrName, sAttrValue);
		if ((int)pxResult == 1) {
			pxResult = hpxCurrentChild->pxElement;
			hpxCurrentChild = hpxCurrentChild->pxNext;
			return pxResult;
			}
		if (pxResult) 	{return pxResult;}
		hpxCurrentChild = hpxCurrentChild->pxNext;
		}
	hpxCurrentChild = hpxChilds;
	return NULL;
	}
//==========================================
xTXMLElement *xTXMLElement::wSearchName (xTType eType, const char *sName)	{
	if (strcmp (sName, hpsName) == 0) if (heType == eType) return this; 
	
	hpxCurrentChild = hpxChilds;
	xTXMLElement *pxResult;
	while (hpxCurrentChild){
		pxResult = hpxCurrentChild->pxElement->wSearchNext(eType, sName);
		if ((int)pxResult == 1) {
			pxResult = hpxCurrentChild->pxElement;
			hpxCurrentChild = hpxCurrentChild->pxNext;
			return pxResult;
			}
		if (pxResult) 	{return pxResult;}
		hpxCurrentChild = hpxCurrentChild->pxNext;
		}
	hpxCurrentChild = hpxChilds;
	return NULL;
	}
//=============================================
xTXMLElement *xTXMLElement::wSearchNameAttr (const char *sName, const char *sAttrName, const char *sAttrValue)	{
	const char* val;
	
    if (strcmp (sName, hpsName) == 0)	{
		if (heType != node) return NULL;
		val = wSearchAttr(sAttrName);
		if (val)
			if(strcmp (val, sAttrValue) == 0) return this;
		}
	
	hpxCurrentChild = hpxChilds;
	xTXMLElement *pxResult;
	while (hpxCurrentChild){
		pxResult = hpxCurrentChild->pxElement->wSearchNextNameAttr(sName, sAttrName, sAttrValue);
		if ((int)pxResult == 1) {
			pxResult = hpxCurrentChild->pxElement;
			hpxCurrentChild = hpxCurrentChild->pxNext;
			return pxResult;
			}
		if (pxResult) 	{return pxResult;}
		hpxCurrentChild = hpxCurrentChild->pxNext;
		}
	hpxCurrentChild = hpxChilds;
	return NULL;
	}
//=============================================
const char *xTXMLElement::wSearchAttr (const char *sName)	{
	xTXMLElementListItem *pxCurrent;
	pxCurrent = hpxAttributes;
	while (pxCurrent){
		if (strcmp (sName, pxCurrent->pxElement->wGetName()) == 0)
			return pxCurrent->pxElement->wGetValue();
		pxCurrent = pxCurrent->pxNext;
		}
	return psNul;
	}
//=============================================
xTXMLElement *xTXMLElement::wSearchName (const char *sName)	{
	return wSearchName (node, sName);
	}
//==============================================			
xTXMLElement *xTXMLElement::wSearchNext (const char *sName){
	return wSearchNext (node, sName);
	}
//==============================================
xTXMLElement *xTXMLElement::wAppend(xTType eType, const char* psName, const char* psValue)	{
	xTXMLElement *pxNewElement = new xTXMLElement(this, eType, psName);
	xTXMLElementListItem *pxNewListItem = new xTXMLElementListItem;
	pxNewListItem->pxElement = pxNewElement;
	pxNewElement->wSetValue(psValue);
	switch (eType)	{
		case node:	{
			if (!hpxChilds)	{
				hpxChilds = pxNewListItem;
				pxNewListItem->pxNext = NULL;
				}
			else	{
				xTXMLElementListItem *pxLastElement = hpxChilds;
				while (pxLastElement->pxNext) pxLastElement = pxLastElement->pxNext;
				pxLastElement->pxNext = pxNewListItem;
				pxNewListItem->pxNext = NULL;
				}
			if (!hpxCurrentChild)	hpxCurrentChild = hpxChilds;
				else	{
					if (hpxCurrentChild == pxNewListItem->pxNext)	
						hpxCurrentChild = pxNewListItem;
					}
			break;
			}	
		case attr:	{
			if (!hpxAttributes) {
				hpxAttributes = pxNewListItem;
				pxNewListItem->pxNext = NULL;
				}
			else	{
				xTXMLElementListItem *pxLastElement = hpxAttributes;
				while (pxLastElement->pxNext) pxLastElement = pxLastElement->pxNext;
				pxLastElement->pxNext = pxNewListItem;
				pxNewListItem->pxNext = NULL;
				}
			if (!hpxCurrentAttr)	hpxCurrentAttr = hpxAttributes;
				else	{
					if (hpxCurrentAttr == pxNewListItem->pxNext)	
						hpxCurrentAttr = pxNewListItem;
					}
			break;
			}
		default:	{
			if (!hpxChilds) {
				hpxChilds = pxNewListItem;
				pxNewListItem->pxNext = NULL;
				}
			else	{
				xTXMLElementListItem *pxLastElement = hpxChilds;
				while (pxLastElement->pxNext) pxLastElement = pxLastElement->pxNext;
				pxLastElement->pxNext = pxNewListItem;
				pxNewListItem->pxNext = NULL;
				}
			if (!hpxCurrentChild)	hpxCurrentChild = hpxChilds;
				else	{
					if (hpxCurrentChild == pxNewListItem->pxNext)	
						hpxCurrentChild = pxNewListItem;
					}
			break;
			}
		}
	return pxNewElement;
	}
//=============================================
void xTXMLElement::wRemove (xTXMLElement *pxElement)	{
	if (hbOnDelete) return;
	switch (pxElement->wType())	{
		
		case node:	{
			if (hpxCurrentChild->pxElement == pxElement) 
				hpxCurrentChild = hpxCurrentChild->pxNext;
			xTXMLElementListItem *pxIter = hpxChilds;
			if (hpxChilds) if (hpxChilds->pxElement == pxElement)	{
				hpxChilds = hpxChilds->pxNext;
				delete pxIter;
				}
			while (pxIter->pxNext)	{
				if (pxIter->pxNext->pxElement == pxElement)	{
					xTXMLElementListItem *pxDel = pxIter->pxNext;
					pxIter->pxNext = pxIter->pxNext->pxNext;
					delete pxDel;
					}
				pxIter = pxIter->pxNext;
				}
			if (!hpxCurrentChild) hpxCurrentChild = hpxChilds;
			break;
			}
			
		case attr:	{
			if (hpxCurrentAttr->pxElement == pxElement) 
				hpxCurrentAttr = hpxCurrentAttr->pxNext;
			xTXMLElementListItem *pxIter = hpxAttributes;
			if (hpxAttributes) if (hpxAttributes->pxElement == pxElement)	{
				hpxAttributes = hpxAttributes->pxNext;
				delete pxIter;
				}
			while (pxIter->pxNext)	{
				if (pxIter->pxNext->pxElement == pxElement)	{
					xTXMLElementListItem *pxDel = pxIter->pxNext;
					pxIter->pxNext = pxIter->pxNext->pxNext;
					delete pxDel;
					}
				pxIter = pxIter->pxNext;
				}
			if (!hpxCurrentAttr) hpxCurrentAttr = hpxAttributes;
			break;
			}
			
		default:	{ //for text
			if (hpxCurrentChild->pxElement == pxElement) 
				hpxCurrentChild = hpxCurrentChild->pxNext;
			xTXMLElementListItem *pxIter = hpxChilds;
			if (hpxChilds) if (hpxChilds->pxElement == pxElement)	{
				hpxChilds = hpxChilds->pxNext;
				delete pxIter;
				}
			while (pxIter->pxNext)	{
				if (pxIter->pxNext->pxElement == pxElement)	{
					xTXMLElementListItem *pxDel = pxIter->pxNext;
					pxIter->pxNext = pxIter->pxNext->pxNext;
					delete pxDel;
					}
				pxIter = pxIter->pxNext;
				}
			if (!hpxCurrentChild) hpxCurrentChild = hpxChilds;
			break;
			}
		}
	}
//============================================
xTXMLElement *xTXMLElement::wFirstAttr()	{
	if (!hpxAttributes) return NULL;
	hpxCurrentAttr = hpxAttributes;
	return hpxCurrentAttr->pxElement;
	}
//--------------------------------------------------------------
xTXMLElement *xTXMLElement::wNextAttr()	{
	if (!hpxCurrentAttr) return NULL;
	if (!hpxCurrentAttr->pxNext) return NULL;
	hpxCurrentAttr = hpxCurrentAttr->pxNext;
	return hpxCurrentAttr->pxElement;
	}
//--------------------------------------------------------------
xTXMLElement *xTXMLElement::wFirstChild()	{
	if (!hpxChilds) return NULL;
	hpxCurrentChild = hpxChilds;
	return hpxCurrentChild->pxElement;
	}
//--------------------------------------------------------------
xTXMLElement *xTXMLElement::wNextChild()	{
	if (!hpxCurrentChild) return NULL;
	if (!hpxCurrentChild->pxNext) return NULL;
	hpxCurrentChild = hpxCurrentChild->pxNext;
	return hpxCurrentChild->pxElement;
	}
//--------------------------------------------------------------
xTXMLElement *xTXMLElement::wParent()	{
	return hpxParent;
	}
//==============================================
inline void xTXMLElement::wSetValue (const char *psValue)	{
	if ((heType != attr) || (!psValue)) return;
	if (hpsValue != psNul) delete hpsValue;
	hpsValue = new char [strlen(psValue)+1];
	strcpy (hpsValue, psValue);
	}
//==============================================
inline void xTXMLElement::wDeleteValue(){
	if ((heType != attr) || (hpsValue == psNul)) return;
	delete hpsValue;
	hpsValue = psNul;
	}	
//=============================================

long xTXMLFile::wReadFromBuffer  (char * psBuffer)	{
     return wReadFromBuffer (psBuffer, xml);
     }
//==================================================
long xTXMLFile::wReadFromBuffer  (char * psBuffer, xTFileType xFileType)	{
	
	heFileType = xFileType;
	#ifdef DEBUG
		{FILE *f = fopen (DEBUGFILE, "w");
		fprintf (f, "%s\n=======================================\n", psBuffer);
		fclose(f);}
	#endif
	
	#ifdef DEBUGSRC
	{FILE *pxFR = fopen (DEBUGSRCFILE, "w");
	fclose (pxFR);}
	#endif
	
	psCurr = psBuffer;
	wParseElementContents (hpxRoot);
	#ifdef DEBUG
		{FILE *f = fopen (DEBUGFILE, "a");
		fprintf(f, "file parsed succesifully!!!\n");
		fclose(f);}
	#endif
	
	}
//=============================================
	
long xTXMLFile::wWriteToFile (FILE *pxFR) {
	lDepth = -1;
	
	xTXMLElement *pxChild = hpxRoot->wFirstChild();
	if (pxChild)	{
		while (pxChild)	{
			wWriteElementContents (pxFR, pxChild);	
			pxChild = hpxRoot->wNextChild();
			}
	}
}	
//==============================================
long xTXMLFile::wParseElementContents (xTXMLElement *pxParent)	{
	
	const char cAttr = 		'=';
	const char cQuote = 	'\"';
	const char cComment = 	'!';
	const char cSpecial = 	'?';
	const char cTagStart = 	'<';
	const char cTagEnd = 	'>';
	const char cElEnd = 	'/';
	const char cEsc = 		' ';
	
	char *psOffStart = psCurr;
	if (heFileType == html)	{
		const char *psParentName = pxParent->wGetName();
	 	char **ppsHtmlUnclosedTags;
		ppsHtmlUnclosedTags = cpsUnclosedTags;
		while (*ppsHtmlUnclosedTags)	{
			if (strcmp (*ppsHtmlUnclosedTags, psParentName) == 0) return 2;
			ppsHtmlUnclosedTags++;
			}
		}

	pgElementStart:
		
		{
		while (*psCurr < 33) if (*psCurr++ == 0) return 1;
		psOffStart = psCurr;
			//omit whitespace
		
		if (*psCurr==cTagStart)	{
			if (*psCurr++ == 0) return 0;	//omit '<'
			psOffStart = psCurr;
		
			switch (*psCurr)	{
			//=================================
				case cSpecial:	{
					//special tag
					if (*psCurr++) return 0;	//omit '?'
					psOffStart = psCurr;
				
					while (*psCurr != cTagEnd) if (*psCurr++ == 0) return 0;
					*(psCurr-1) = 0;
					xTXMLElement *pxNewElement = pxParent->wAppend(xTXMLElement::special, psOffStart, NULL);
					
					#ifdef DEBUG
						{FILE *f = fopen (DEBUGFILE, "a");
						fprintf (f, "got special %s\n", pxNewElement->wGetName());
						fclose(f);}
					#endif
					
					if (*psCurr++ == 0) return 0;	//step onto the next character after '>'
					goto pgElementStart;
					}
			//=================================		
				case cComment:	{
					//comment
					if (*psCurr++ == 0) return 0;	//omit '!'
					psOffStart = psCurr;
					
					while (*psCurr != cTagEnd) if (*psCurr++ == 0) return 0;
					*psCurr = 0;
					xTXMLElement *pxNewElement = pxParent->wAppend(xTXMLElement::comment, psOffStart, NULL);
					*psCurr = cTagEnd;
					
					#ifdef DEBUG
						{FILE *f = fopen (DEBUGFILE, "a");
						fprintf (f, "got comment %s\n>>\n", pxNewElement->wGetName());
						fclose(f);}
					#endif
					
					if (*psCurr++ == 0) return 0;	//step onto the next character after '>'
					goto pgElementStart;
					}
			//=================================	
				case cElEnd:	{
					//element end
					if (*psCurr++ == 0) return 0;	//omit '/'
					psOffStart = psCurr;
					
					while (*psCurr != cTagEnd) if (*psCurr++ == 0) return 0;
					*psCurr = 0;
					if (strcmp (psOffStart, pxParent->wGetName())==0) {
						*psCurr = cTagEnd;
						if (*psCurr++ == 0) return 0;	//step onto the next character afer '>'
						
						#ifdef DEBUG
							FILE *f = fopen (DEBUGFILE, "a");
							fprintf (f, "got end %s\n>>\n", pxParent->wGetName());
							fclose(f);
						#endif
						
						goto pgReturn; 
						}
						
					#ifdef DEBUG
						{FILE *f = fopen (DEBUGFILE, "a");
						fprintf (f, "end miss: bad xml \n\toff: %s\n\tel:  %s\n>>\n", psOffStart, pxParent->wGetName());
						fclose(f);}
					#endif
					*psCurr = cTagEnd;
					if (*psCurr++ == 0) return 0; //step onto the next character afer '>'
					goto pgElementStart;;
					}
			//=================================		
				default:	{
					//simple tag
					
					while ((*psCurr!=cTagEnd)&&(*psCurr!=cEsc)) if (*psCurr++ == 0) return 0;
					
					if (*psCurr == cTagEnd)
							{	
							//the tag without any attributes
							if (*(psCurr-1) == cElEnd)	{
								// got enclosed tag
								*(psCurr-1) = 0;
								#ifdef DEBUG
									{FILE *f = fopen (DEBUGFILE, "a");
									fprintf (f, "<<comparing tags %s ; %s\n", psOffStart, pxParent->wGetName());
									fclose(f);}
								#endif	
								if (strcmp (psOffStart, pxParent->wGetName()) == 0) {	//got tag end
									if (*psCurr++ == 0) return 0;
									#ifdef DEBUG
										{FILE *f = fopen (DEBUGFILE, "a");
										fprintf (f, "<result: match\n");
										fclose(f);}
									#endif	
									goto pgReturn;
									}
								
								#ifdef DEBUG
									{FILE *f = fopen (DEBUGFILE, "a");
									fprintf (f, "<result: differs\n");
									fclose(f);}
								#endif
								
								xTXMLElement *pxNewElement = pxParent->wAppend (xTXMLElement::node, psOffStart, NULL);
								*(psCurr-1) = cElEnd;
								
								#ifdef DEBUG
									{FILE *f = fopen (DEBUGFILE, "a");
									fprintf (f, "<<got tag %s\n", pxNewElement->wGetName());
									fclose(f);}
								#endif
								
								if (*psCurr++ == 0) return 0;	//step onto the next character afer '>'
								}
								else	{
								//got tag in format <word>
								*psCurr = 0;
								xTXMLElement *pxNewElement = pxParent->wAppend(xTXMLElement::node, psOffStart, NULL);
								
								#ifdef DEBUG
									{FILE *f = fopen (DEBUGFILE, "a");
									fprintf (f, "<<got tag %s\n", pxNewElement->wGetName());
									fclose(f);}
								#endif
								
								*psCurr = cTagEnd;
								if (*psCurr++ == 0) return 0;	//step onto the next character afer '>'
								#ifdef DEBUG
									{FILE *f = fopen (DEBUGFILE, "a");
									fprintf (f, "parsing content of %s\n", pxNewElement->wGetName());
									fclose(f);}
								#endif
								switch (wParseElementContents (pxNewElement))	{
									case 0: //erroneus eof		
										{return 0;};
									case 1:	//good eof
										{return 1;};
									case 2:	//only end of element
										{goto pgElementStart;}
									}
								}
							}
						else	{
							//got esc: tag is with attributes attached
							*psCurr = 0;
							xTXMLElement *pxNewElement = pxParent->wAppend(xTXMLElement::node, psOffStart, NULL);
							*psCurr = cEsc;
							
							#ifdef DEBUG
									{FILE *f = fopen (DEBUGFILE, "a");
									fprintf (f, "<<got tag %s\n", pxNewElement->wGetName());
									fclose(f);}
							#endif
							
							while (*psCurr < 33) if (*psCurr++ == 0) return 0; 	//omit whitespace
							psOffStart = psCurr;
								
							while (1){
								while (*psCurr!=cAttr) if (*psCurr++ == 0) return 0;
								char *psOffEnd = psCurr;
								if (*psCurr++ == 0) return 0; 	//step onto the next character after '='
								char *psAttrStart = psCurr;
								
								if (*psCurr==cQuote)	{
									psAttrStart++;
									if (*psCurr++ == 0) return 0;
									while (*psCurr!=cQuote) if (*psCurr++ == 0) return 0;
									
									*psCurr = 0;
									*psOffEnd = 0;
									xTXMLElement * pxNewAttr = pxNewElement->wAppend(xTXMLElement::attr, psOffStart, psAttrStart);
									
									#ifdef DEBUG
										{FILE *f = fopen (DEBUGFILE, "a");
										fprintf (f, "got attr %s :: %s\n", pxNewAttr->wGetName(), pxNewAttr->wGetValue());
										fclose(f);}
									#endif
									
									*psCurr = cQuote;
									*psOffEnd = cAttr;
									if (*psCurr++ == 0) return 0;	//step onto the next character after '"'
									}
									
									else	{
									while ((*psCurr!=cEsc)&&(*psCurr!=cElEnd)&&(*psCurr!=cTagEnd)) psCurr++;
									char c = *psCurr;
									*psCurr = 0;
									*psOffEnd = 0;
									xTXMLElement * pxNewAttr = pxNewElement->wAppend(xTXMLElement::attr, psOffStart, psAttrStart);
									
									#ifdef DEBUG
										{FILE *f = fopen (DEBUGFILE, "a");
										fprintf (f, "got attr %s :: %s\n", pxNewAttr->wGetName(), pxNewAttr->wGetValue());
										fclose(f);}
									#endif
									
									*psCurr = c;
									*psOffEnd = cAttr; 
									}
								while (*psCurr < 33) if (*psCurr++ == 0) return 0; 	//omit whitespace
								psOffStart = psCurr;
								
								if (*psCurr==cElEnd) {if (*(psCurr+1)==cTagEnd) psCurr+=2;} //step after '/>'
									else if (*psCurr==cTagEnd)	{
										if (*psCurr++ == 0) return 0;
										#ifdef DEBUG
											{FILE *f = fopen (DEBUGFILE, "a");
											fprintf (f, "parsing content of %s\n", pxNewElement->wGetName());
											fclose(f);}
										#endif
										switch (wParseElementContents (pxNewElement))	{
											case 0: //erroneus eof		
												{return 0;};
											case 1:	//good eof
												{return 1;};
											case 2:	//only end of element
												{goto pgElementStart;}
											}
			
									}
								}	
										
							}
						
					}
				}
			
			}
			else	{
			//text
			while ((*psCurr!=cTagStart)&&(*psCurr!=0)) psCurr++; //omit text
			if (*psCurr == 0)	{
				xTXMLElement *pxNewElement = pxParent->wAppend(xTXMLElement::text, psOffStart, NULL);
				
				#ifdef DEBUG
					{FILE *f = fopen (DEBUGFILE, "a");
					fprintf (f, "got text at end %s\n", pxNewElement->wGetName());
					fclose(f);}
				#endif
				
				return 1;
				}
			else	{
				*psCurr = 0;
				xTXMLElement *pxNewElement = pxParent->wAppend(xTXMLElement::text, psOffStart, NULL);
				
				#ifdef DEBUG
					{FILE *f = fopen (DEBUGFILE, "a");
					fprintf (f, "got text %s\n", pxNewElement->wGetName());
					fclose(f);}
				#endif
				
				*psCurr = cTagStart;
				goto pgElementStart;
				}
			}
		}
pgReturn:
	return 2;
	}	
//===========================================================
long xTXMLFile::wWriteElementContents (FILE *pxFR, xTXMLElement *pxParent)	{
	
	lDepth++;
	#ifdef DEBUG
		{FILE *f = fopen (DEBUGFILE, "a");
		fprintf (f, "printing %s::%s\n", pxParent->wGetName(), pxParent->wGetValue());
		fclose(f);}
	#endif
	switch (pxParent->wType())	{
		case xTXMLElement::node:	{
			
			for (long i = 0; i<lDepth; i++) fprintf (pxFR, "\t"); 
			fprintf (pxFR, "<%s", pxParent->wGetName());
			xTXMLElement *pxAttr = pxParent->wFirstAttr();
			
			if (pxAttr) fprintf(pxFR, " ");
			while (pxAttr)	{
				
				#ifdef DEBUG
					{FILE *f = fopen (DEBUGFILE, "a");
					fprintf (f, "printing attr%s::%s\n", pxAttr->wGetName(), pxAttr->wGetValue());
					fclose(f);}
				#endif
				
				fprintf(pxFR, "%s=\"%s\" ", pxAttr->wGetName(), pxAttr->wGetValue());
				pxAttr = pxParent->wNextAttr();
				}
			
			if (heFileType == html)	{
				const char *psParentName = pxParent->wGetName();
				char **ppsHtmlUnclosedTags;
				char **ppsHtmlClosedTags;
				ppsHtmlUnclosedTags = cpsUnclosedTags;
				ppsHtmlClosedTags = cpsClosedTags;
				while (*ppsHtmlUnclosedTags)	{
					if (strcmp (*ppsHtmlUnclosedTags, psParentName) == 0)	{
						fprintf(pxFR, ">\n");
						goto pgReturn;
						}
					ppsHtmlUnclosedTags++;
					}
					
				while (*ppsHtmlClosedTags)	{
					if (strcmp (*ppsHtmlClosedTags, psParentName) == 0)	{
						fprintf(pxFR, "></%s>\n", psParentName);
						goto pgReturn;
						}
					ppsHtmlClosedTags++;
					}
				}
			xTXMLElement *pxChild = pxParent->wFirstChild();
			if (pxChild)	{
			
				fprintf(pxFR, ">\n");
				while (pxChild)	{
					wWriteElementContents (pxFR, pxChild);	
					pxChild = pxParent->wNextChild();
					}
				for (long i = 0; i<lDepth; i++) fprintf (pxFR, "\t");
				fprintf (pxFR, "</%s>\n", pxParent->wGetName());
				}
				else fprintf(pxFR, "/>\n");
			break;
			}
		
		case  xTXMLElement::attr: break;
		
		case  xTXMLElement::text:	{
			for (long i = 0; i<lDepth; i++) fprintf (pxFR, "\t");
			fprintf (pxFR, "%s\n", pxParent->wGetName());
			break;
			}
			
		case  xTXMLElement::special:	{
			for (long i = 0; i<lDepth; i++) fprintf (pxFR, "\t");
			fprintf (pxFR, "<?%s?>\n", pxParent->wGetName());
			break;
			}	
			
		case  xTXMLElement::comment:	{
			for (long i = 0; i<lDepth; i++) fprintf (pxFR, "\t");
			fprintf (pxFR, "<!%s>\n", pxParent->wGetName());
			break;
			}	
		}
	pgReturn:
	pxParent->wFirstChild();
	lDepth--;
	return 1;
	}
	
