/* 	xmlfile.h
Part of xml file parser

	version 2.2
	
	2.2 added advanced search both by tag and the attributes included
	2.1 added search by attributes in xTXMLElement
	2.0 added search in xTXMLElement
	1.2 corrected minor errors with html
	1.1 added html file support
	1.0 added xml file support
	0.2 corrected errors with new xTXMLElements
	0.1 initial version
	
	Written by Povilas Kanapickas [exec]

	This software comes "AS IS" therefore the author is NOT responsible for any 
	losses or damage it can cause.

	Published under GPL (see http://www.gnu.org/gpl)
*/

#ifndef XMLFILE_H
#define XMLFILE_H

#include <fstream.h>

//#define DEBUG
//#define DEBUGSRC
#ifdef DEBUG
#define DEBUGFILE "err.txt"
#endif

#ifdef DEBUGSRC
#define DEBUGSRCFILE "err2.txt"
#endif


class xTXMLElement;

typedef struct	xTXMLElementListItem{
	xTXMLElement *pxElement;
	xTXMLElementListItem *pxNext;
	} xTXMLElementListItem;

//============================================
class xTXMLElement	{
	public:
		typedef enum {
			node, attr, text, special, comment
			}xTType;
	
	private:	
		char * psNul;
		xTType heType;
		char *hpsName;

		//if type: node
		xTXMLElementListItem *hpxAttributes;
		xTXMLElementListItem *hpxCurrentAttr;
		
		xTXMLElementListItem *hpxChilds;
		xTXMLElementListItem *hpxCurrentChild;
		
		xTXMLElement *hpxParent;
		//if type: attr
		char *hpsValue;
		bool hbOnDelete;
	public: //do not use
		xTXMLElement(xTXMLElement *pxParent, xTType heType, const char* psName);
		void wRemove (xTXMLElement *pxElement);
		
	public:
		
		xTType wType()	{return heType;};
		
		xTXMLElement *wFirstAttr();
		xTXMLElement *wNextAttr();
		
		xTXMLElement *wFirstChild();
		xTXMLElement *wNextChild();
		xTXMLElement *wParent();
		
		const char *wGetName()	{ return hpsName;};
		int wGetInt();
		const char *wGetValue()	{ if (heType == attr) return hpsValue;
												else  return psNul;};
		
		xTXMLElement *wSearchName (xTType eType, const char *sName);
		xTXMLElement *wSearchNext (xTType eType, const char *sName);
		
		xTXMLElement *wSearchName (const char *sName);
		xTXMLElement *wSearchNext (const char *sName);
		
		xTXMLElement *xTXMLElement::wSearchNameAttr (const char *sName, const char *sAttrName, const char *sAttrValue);
		xTXMLElement *xTXMLElement::wSearchNextNameAttr (const char *sName, const char *sAttrName, const char *sAttrValue);
		
		const char *wSearchAttr (const char *sName);
		inline void wSetValue (const char *psValue);
		inline void wDeleteValue();
		
		xTXMLElement *wAppend(xTType eType, const char* psName, const char* psValue);
		
		xTXMLElement();
		~xTXMLElement();
	};


class xTXMLFile	{
	public: 
		typedef enum {html, xml}xTFileType;
	
	private:
		xTXMLElement *hpxRoot;
		char *psCurr;
		xTFileType heFileType;
		char** cpsUnclosedTags;
		char** cpsClosedTags;
		long wParseElementContents (xTXMLElement *pxParent);
		long wWriteElementContents (FILE *pxFR, xTXMLElement *pxParent);
		long lDepth;
	public:
	
		xTXMLElement *wRoot()	{return hpxRoot;};
		
		long wReadFromBuffer (char * psBuffer, xTFileType xFileType);
		long wReadFromBuffer (char * psBuffer);
		long wWriteToFile (FILE *pxFR);
		
		void wFree()	{if(hpxRoot) delete hpxRoot;
						 hpxRoot = new xTXMLElement;};
		
		xTXMLFile()		{hpxRoot = new xTXMLElement;
						wFree();
						cpsUnclosedTags = new char*[20];
						cpsUnclosedTags[0] = "input";
						cpsUnclosedTags[1] = "option";
						cpsUnclosedTags[2] = "col";
						cpsUnclosedTags[3] = "hr";
						cpsUnclosedTags[4] = "param";
						cpsUnclosedTags[5] = "area";
						cpsUnclosedTags[6] = "br";
						cpsUnclosedTags[7] = "img";
						cpsUnclosedTags[8] = "isindex";
						cpsUnclosedTags[9] = "meta";
						cpsUnclosedTags[10] = "basefont";
						cpsUnclosedTags[11] = "link";
						cpsUnclosedTags[12] = "base";
						cpsUnclosedTags[13] = NULL;
						cpsClosedTags = new char*[5];
						cpsClosedTags[0] = "script";
						cpsClosedTags[1] = NULL;};
						
		~xTXMLFile()	{if(hpxRoot) delete hpxRoot;
						delete[]cpsClosedTags;
						delete[]cpsUnclosedTags;};
	};	
		
#endif
